"""A list of databricks reserved keywords.

https://docs.databricks.com/sql/language-manual/sql-ref-reserved-words.html
"""

RESERVED_KEYWORDS = [
    "ANTI",
    "CROSS",
    "EXCEPT",
    "FULL",
    "INNER",
    "INTERSECT",
    "JOIN",
    "LATERAL",
    "LEFT",
    "MINUS",
    "NATURAL",
    "ON",
    "RIGHT",
    "SEMI",
    "UNION",
    "USING",
]

UNRESERVED_KEYWORDS = [
    "AUTO",
    "CATALOG",
    "CDC",
    "CLONE",
    "COMPENSATION",
    "CRON",
    "DAY",
    "DEEP",
    "DAYS",
    "ENFORCED",
    "EVOLUTION",
    "FEATURE",
    "FILTER",
    "FLOW",
    "HOUR",
    "HOURS",
    "IDENTIFIER",
    "MANAGED",
    "MASK",
    "METRICS",
    "MOST",
    "NORELY",
    "OPTIMIZATION",
    "OPTIMIZE",
    "PREDICTIVE",
    "PRIVATE",
    "PROVIDER",
    "PYTHON",
    "RECIPIENT",
    "REFRESH",
    "RELY",
    "ROW",
    "SCHEDULE",
    "SHALLOW",
    "SOURCE",
    "SQL",
    "TAG",
    "TAGS",
    "TARGET",
    "TIMESERIES",
    "TRIGGER",
    "VOLUME",
    "VOLUMES",
    "WEEK",
    "WEEKS",
    "YAML",
    "ZORDER",
]
