"""A list of Oracle SQL keywords.

https://docs.oracle.com/en/database/oracle/oracle-database/23/sqlrf/
Oracle-SQL-Reserved-Words.html
"""

oracle_reserved_keywords = """
ACCESS
ADD
ALL
ALTER
AND
ANY
AS
ASC
AUDIT
BETWEEN
BY
CHAR
CHECK
CLUSTER
COLUMN
COLUMN_VALUE
COMMENT
COMPRESS
CONNECT
CONNECT_BY_ROOT
CONSTRAINT
CREATE
CURRENT
DATE
DECIMAL
DEFAULT
DEFINITION
DELETE
DELETING
DESC
DISABLE
DISTINCT
DROP
ELSE
ENABLE
EXCLUSIVE
EXECUTE
EXISTS
FILE
FLOAT
FOR
FORCE
FROM
GRANT
GROUP
HAVING
IDENTIFIED
IMMEDIATE
IN
INCREMENT
INDEX
INDEXTYPE
INITIAL
INSERT
INSERTING
INTEGER
INTERSECT
INTO
INVISIBLE
IS
LEVEL
LIKE
LOCK
LOGGING
LONG
LOOP
MAXEXTENTS
MINUS
MLSLABEL
MODE
MODIFY
MONITORING
MULTISET
NESTED_TABLE_ID
NOAUDIT
NOCOMPRESS
NOLOGGING
NOMONITORING
NOREVERSE
NOT
NOWAIT
NULL
NUMBER
OF
OFFLINE
ON
ONLINE
OPTION
OR
ORDER
OVERFLOW
PARAMETERS
PCTFREE
PIVOT
PRIOR
PRIVATE
PROMPT
PUBLIC
RAW
REBUILD
RENAME
RESOURCE
REVERSE
REVOKE
ROW
ROWID
ROWNUM
ROWS
SELECT
SESSION
SET
SHARE
SIBLINGS
SIZE
SMALLINT
START
SUCCESSFUL
SYNONYM
SYSDATE
TABLE
THEN
TO
TRIGGER
UID
UNION
UNIQUE
UNPIVOT
UNUSABLE
UPDATE
UPDATING
USER
VALIDATE
VALUES
VARCHAR
VARCHAR2
VIEW
VISIBLE
WHEN
WHENEVER
WHERE
WITH
"""

oracle_unreserved_keywords = """
ABSENT
ACCESSIBLE
ACTIVE
ADMINISTER
ADVANCED
ADVISE
ADVISOR
ANALYTIC
ARCHIVAL
ARCHIVE
AUTHENTICATED
AUTHID
AUTO
BASIC
BECOME
BITMAP
BODY
BUFFER_POOL
BULK
BULK_EXCEPTIONS
BULK_ROWCOUNT
BYTE
CAPACITY
CELL_FLASH_CACHE
COLLECT
COMMITTED
COMPILE
COMPOUND
CONSTANT
CONSTRAINTS
CONTAINER
CONTEXT
CREATION
CRITICAL
CROSSEDITION
CURRVAL
CURSOR
DBA_RECYCLEBIN
DBTIMEZONE
DDL
DEBUG
DEFERRED
DELEGATE
DENSE_RANK
DIGEST
DIMENSION
DIRECTIVE
DIRECTORIES
DIRECTORY
DISTRIBUTE
DML
DUPLICATE
EDITION
EDITIONABLE
EDITIONING
EDITIONS
ELSIF
EMPTY
ERROR
ERRORS
EVALNAME
EXEMPT
EXPIRE
EXTERNALLY
FINE
FLASHBACK
FLASH_CACHE
FOLLOWS
FORALL
FREELIST
FREELISTS
GLOBALLY
GROUPS
GUARD
HIERARCHY
HIGH
HTTP
INDICES
INHERITANY
INITRANS
INMEMORY
ISOLATION_LEVEL
ISOPEN
JAVA
JOB
KEEP
LIBRARY
LINK
LOCKDOWN
LOCKED
LOCKING
LOG
LOGMINING
LOOP
LOW
MAXSIZE
MAXTRANS
MEASURE
MEDIUM
MEMCOMPRESS
MINEXTENTS
MINING
MOVEMENT
MUTABLE
NESTED
NEXTVAL
NOCOPY
NOMAXVALUE
NOMINVALUE
NONE
NONEDITIONABLE
NOPARALLEL
NOROWDEPENDENCIES
NOSORT
NOTFOUND
NOTHING
OID
OLTP
OPTIMAL
ORA_ROWSCN
OUTLINE
PACKAGE
PAIRS
PARALLEL
PARALLEL_ENABLE
PARENT
PCTINCREASE
PCTUSED
PERSISTABLE
PIPELINED
PLUGGABLE
POLYMORPHIC
PRAGMA
PRECEDES
PRIORITY
PRIVILEGE
PROFILE
PROGRAM
PROPERTY
QUERY
QUOTA
RAISE
RECORD
RECYCLE
REDACTION
REDEFINE
REFRESH
REJECT
RELIES_ON
REMOTE
RESTRICTED
RESULT_CACHE
RESUMABLE
RETURNING
REUSE
REVERSE
REWRITE
ROWDEPENDENCIES
ROWTYPE
SCHEDULER
SEGMENT
SERIALIZABLE
SERVICE
SESSION_USER
SHARD
SHARD_ENABLE
SHARED
SHARING
SIGN
SKIP
SPECIFICATION
SQL_MACRO
STORAGE
STORE
SUBPARTITION
SYNC
SYSGUID
TIMEOUT
TIME_ZONE
UNLIMITED
VARRAY
VISIBILITY
WAIT
"""
